import numpy as np
from pathlib import Path
from multiprocessing import shared_memory

import example_module

shm: shared_memory.SharedMemory | None = None


def segment(imagePath: Path | str):
    global shm
    # Segment the image with example_module.py
    masks, flows, styles, diams = example_module.segment(imagePath)
    # Create the shared memory
    shm = shared_memory.SharedMemory(create=True, size=masks.nbytes)
    # Create a NumPy array backed by shared memory
    masksShared = np.ndarray(masks.shape, dtype=masks.dtype, buffer=shm.buf)
    # Copy the masks into the shared memory
    masksShared[:] = masks[:]
    # Return the shape, dtype and shared memory name to recreate the numpy array on the other side
    return masks.shape, masks.dtype, shm.name


def clean():
    global shm
    if shm is None:
        return
    # Clean up the shared memory in this process
    shm.close()
    # Free and release the shared memory block
    shm.unlink()



class ExecutionException(Exception):
    """Exception raised when the environment raises an error when executing the requested function.

    Attributes:
            message: explanation of the error
    """

    def __init__(self, message):
        super().__init__(message)
        self.exception = message["exception"] if "exception" in message else None
        self.traceback = message["traceback"] if "traceback" in message else None


class IncompatibilityException(Exception):
    pass



from unittest.mock import MagicMock

import pytest

from wetlands.environment_manager import EnvironmentManager


# ---- conda_path Tests ----


def test_environment_manager_conda_path(tmp_path_factory):
    """Check that an exception is raised if the provided CondaPath contains the name of another conda manager."""

    dummy_conda_path = tmp_path_factory.mktemp("path_containing_pixi_and_micromamba").resolve()
    wetlands_instance_path = tmp_path_factory.mktemp("wetlands_instance").resolve()

    # With manager="auto", it should raise because the path is ambiguous
    with pytest.raises(Exception, match="must contain either"):
        manager = EnvironmentManager(
            wetlandsInstancePath=wetlands_instance_path,
            condaPath=tmp_path_factory.mktemp("random_path").resolve(),
            manager="auto",
        )

    # With explicit manager, it should not raise based on path
    manager = EnvironmentManager(
        wetlandsInstancePath=wetlands_instance_path, condaPath=dummy_conda_path, manager="pixi"
    )
    assert manager is not None

    manager = EnvironmentManager(
        wetlandsInstancePath=wetlands_instance_path, condaPath=dummy_conda_path, manager="micromamba"
    )
    assert manager is not None


# ---- setCondaPath Tests ----


class TestSetCondaPath:
    def test_set_conda_path_updates_settings(self, tmp_path_factory, monkeypatch):
        """Test that setCondaPath updates the settings manager"""
        dummy_micromamba_path = tmp_path_factory.mktemp("conda_root")
        wetlands_instance_path = tmp_path_factory.mktemp("wetlands_instance")
        main_env_path = dummy_micromamba_path / "envs" / "main_test_env"

        monkeypatch.setattr(EnvironmentManager, "installConda", MagicMock())

        manager = EnvironmentManager(
            wetlandsInstancePath=wetlands_instance_path,
            condaPath=dummy_micromamba_path,
            manager="micromamba",
            mainCondaEnvironmentPath=main_env_path,
        )

        new_path = tmp_path_factory.mktemp("new_conda")

        mock_set_path = MagicMock()
        monkeypatch.setattr(manager.settingsManager, "setCondaPath", mock_set_path)

        manager.setCondaPath(new_path, usePixi=True)

        mock_set_path.assert_called_once_with(new_path, True)


# ---- setProxies Tests ----


class TestSetProxies:
    def test_set_proxies_calls_settings_manager(self, tmp_path_factory, monkeypatch):
        """Test that setProxies delegates to settings manager"""
        dummy_micromamba_path = tmp_path_factory.mktemp("conda_root")
        wetlands_instance_path = tmp_path_factory.mktemp("wetlands_instance")
        main_env_path = dummy_micromamba_path / "envs" / "main_test_env"

        monkeypatch.setattr(EnvironmentManager, "installConda", MagicMock())

        manager = EnvironmentManager(
            wetlandsInstancePath=wetlands_instance_path,
            condaPath=dummy_micromamba_path,
            manager="micromamba",
            mainCondaEnvironmentPath=main_env_path,
        )

        proxies = {"http": "http://proxy.example.com:8080", "https": "https://proxy.example.com:8443"}

        mock_set_proxies = MagicMock()
        monkeypatch.setattr(manager.settingsManager, "setProxies", mock_set_proxies)

        manager.setProxies(proxies)

        mock_set_proxies.assert_called_once_with(proxies)



from pathlib import Path
from wetlands.environment_manager import EnvironmentManager
import urllib.request


def initialize():
    # Initialize the environment manager
    # Wetlands will store logs and state in the wetlandsInstancePath (defaults to "wetlands/")
    # Pixi/Micromamba will be installed in wetlandsInstancePath/pixi by default
    environmentManager = EnvironmentManager()

    # Create and launch an isolated Conda environment named "cellpose"
    env = environmentManager.create("cellpose", {"conda": ["cellpose==3.1.0"]})
    env.launch()

    # Download example image from cellpose
    imagePath = Path("cellpose_img02.png")
    imageUrl = "https://www.cellpose.org/static/images/img02.png"

    with urllib.request.urlopen(imageUrl) as response:
        imageData = response.read()

    with open(imagePath, "wb") as handler:
        handler.write(imageData)

    segmentationPath = imagePath.parent / f"{imagePath.stem}_segmentation.png"
    return imagePath, segmentationPath, env


if __name__ == "__main__":
    # Initialize: create the environment manager, the Cellpose conda environment, and download the image to segment
    imagePath, segmentationPath, env = initialize()

    # Import example_module in the environment
    exampleModule = env.importModule("example_module.py")
    # exampleModule is a proxy to example_module.py in the environment,
    # calling exampleModule.function_name(args) will run env.execute(module_name, function_name, args)
    diameters = exampleModule.segment(str(imagePath), str(segmentationPath))

    # Or use env.execute() directly to call a function in a module
    # diameters = env.execute("example_module.py", "segment", (imagePath, segmentationPath))

    # Alternatively, use env.runScript() to run an entire Python script
    # env.runScript("script.py", args=(str(imagePath), str(segmentationPath)))

    print(f"Found diameters of {diameters} pixels.")

    # Clean up and exit the environment
    env.exit()



import pytest
from pathlib import Path
from unittest.mock import MagicMock, patch
from types import ModuleType
from wetlands.environment import Environment


class DummyEnvironment(Environment):
    def launch(self, additionalActivateCommands={}, logOutputInThread=True):
        pass

    def execute(self, modulePath, function, args=[], kwargs={}):
        return f"Executed {function} in {modulePath} with args {args} and kwargs {kwargs}"


@pytest.fixture
def mock_environment_manager():
    return MagicMock()


@pytest.fixture
def dummy_env(mock_environment_manager):
    return DummyEnvironment("test_env", Path("/tmp/test_env"), mock_environment_manager)


@patch("sys.path", new=[])
@patch("wetlands.environment.import_module")
def test_importModule(mock_import_module, dummy_env):
    mock_mod = ModuleType("test_mod")
    mock_import_module.return_value = mock_mod

    module = dummy_env._importModule("/path/to/test_mod.py")
    assert module == mock_mod
    assert "test_mod" in dummy_env.modules
    assert dummy_env.modules["test_mod"] == mock_mod


@patch("wetlands.environment.Environment._importModule")
@patch("wetlands.environment.Environment._listFunctions")
def test_importModule_creates_fake_module(mock_listFunctions, mock_importModule, dummy_env):
    mock_mod = MagicMock()
    mock_importModule.return_value = mock_mod
    mock_listFunctions.return_value = ["func1", "func2"]

    fake_module = dummy_env.importModule("/path/to/test_mod.py")

    assert hasattr(fake_module, "func1")
    assert hasattr(fake_module, "func2")

    result = fake_module.func1("value1")
    assert result == "Executed func1 in /path/to/test_mod.py with args ('value1',) and kwargs {}"

    result = fake_module.func2("value2", arg_name="arg_value")
    assert result == "Executed func2 in /path/to/test_mod.py with args ('value2',) and kwargs {'arg_name': 'arg_value'}"


def test_exit(dummy_env, mock_environment_manager):
    dummy_env.exit()
    mock_environment_manager._removeEnvironment.assert_called_once_with(dummy_env)



import sys
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from wetlands.environment_manager import EnvironmentManager
from wetlands._internal.dependency_manager import Dependencies

# --- Fixtures (shared from conftest if needed) ---

conda_list_json = """
[
    {
        "base_url": "https://repo.anaconda.com/pkgs/main",
        "build_number": 1,
        "build_string": "h18a0788_1",
        "channel": "pkgs/main",
        "dist_name": "zlib-1.2.13-h18a0788_1",
        "name": "zlib",
        "platform": "osx-arm64",
        "version": "1.2.13"
    },
    {
        "base_url": "https://repo.anaconda.com/pkgs/main",
        "build_number": 0,
        "build_string": "py312h1a4646a_0",
        "channel": "pkgs/main",
        "dist_name": "zstandard-0.22.0-py312h1a4646a_0",
        "name": "zstandard",
        "platform": "osx-arm64",
        "version": "0.22.0"
    },
    {
        "base_url": "https://repo.anaconda.com/pkgs/main",
        "build_number": 2,
        "build_string": "hd90d995_2",
        "channel": "pkgs/main",
        "dist_name": "zstd-1.5.5-hd90d995_2",
        "name": "zstd",
        "platform": "osx-arm64",
        "version": "1.5.5"
    }
]
    """.splitlines()


@pytest.fixture
def mock_command_executor(monkeypatch):
    """Mocks the CommandExecutor methods."""
    import subprocess

    mock_execute = MagicMock(spec=subprocess.Popen)
    mock_execute_output = MagicMock(return_value=["output line 1", "output line 2"])

    mocks = {
        "executeCommands": mock_execute,
        "executeCommandsAndGetOutput": mock_execute_output,
    }
    return mocks


@pytest.fixture
def environment_manager_fixture(tmp_path_factory, mock_command_executor, monkeypatch):
    """Provides an EnvironmentManager instance with mocked CommandExecutor."""
    dummy_micromamba_path = tmp_path_factory.mktemp("conda_root")
    wetlands_instance_path = tmp_path_factory.mktemp("wetlands_instance")
    main_env_path = dummy_micromamba_path / "envs" / "main_test_env"

    monkeypatch.setattr(EnvironmentManager, "installConda", MagicMock())

    manager = EnvironmentManager(
        wetlandsInstancePath=wetlands_instance_path,
        condaPath=dummy_micromamba_path,
        manager="micromamba",
        mainCondaEnvironmentPath=main_env_path,
    )

    monkeypatch.setattr(manager.commandExecutor, "executeCommands", mock_command_executor["executeCommands"])
    monkeypatch.setattr(
        manager.commandExecutor, "executeCommandsAndGetOutput", mock_command_executor["executeCommandsAndGetOutput"]
    )

    monkeypatch.setattr(manager, "environmentExists", MagicMock(return_value=False))

    return manager, mock_command_executor["executeCommandsAndGetOutput"], mock_command_executor["executeCommands"]


# ---- _dependenciesAreInstalled Tests ----


def test_environment_validates_requirements_main_env_python_mismatch(environment_manager_fixture):
    manager, mock_execute_output, _ = environment_manager_fixture
    # Ensure the version string format causes a mismatch
    different_py_version = "99.99"
    assert not sys.version.startswith(different_py_version)

    dependencies: Dependencies = {"python": f"={different_py_version}"}

    result = manager._environmentValidatesRequirements(manager.mainEnvironment, dependencies)

    assert result is False
    mock_execute_output.assert_not_called()


def test_environment_validates_requirements_main_env_empty_deps(environment_manager_fixture):
    manager, mock_execute_output, _ = environment_manager_fixture
    dependencies: Dependencies = {}

    result = manager._environmentValidatesRequirements(manager.mainEnvironment, dependencies)

    assert result is True
    mock_execute_output.assert_not_called()


def test_environment_validates_requirements_main_env_no_path_conda_fails(environment_manager_fixture):
    manager, mock_execute_output, _ = environment_manager_fixture
    manager.mainEnvironment.path = None
    dependencies: Dependencies = {"conda": ["some_package"]}

    result = manager._environmentValidatesRequirements(manager.mainEnvironment, dependencies)

    assert result is False
    mock_execute_output.assert_not_called()


def test_environment_validates_requirements_main_env_no_path_pip_uses_metadata(environment_manager_fixture):
    manager, mock_execute_output, _ = environment_manager_fixture
    manager.mainEnvironment.path = None
    dependencies: Dependencies = {"pip": ["pytest"]}

    result = manager._environmentValidatesRequirements(manager.mainEnvironment, dependencies)

    # This depends on whether 'pytest' is ACTUALLY available via metadata in the test env
    import importlib.metadata

    try:
        importlib.metadata.version("pytest")
        assert result is True
    except importlib.metadata.PackageNotFoundError:
        assert result is False

    mock_execute_output.assert_not_called()


# ---- _checkRequirement Tests ----


def test_check_requirement_no_version_specified(environment_manager_fixture):
    """Test matching package without version constraint."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.0", "kind": "pypi"}]

    result = manager._checkRequirement("numpy", "pip", installed_packages)

    assert result is True


def test_check_requirement_exact_version_match(environment_manager_fixture):
    """Test exact version matching with ==."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.0", "kind": "pypi"}]

    result = manager._checkRequirement("numpy==1.20.0", "pip", installed_packages)

    assert result is True


def test_check_requirement_exact_version_no_match(environment_manager_fixture):
    """Test exact version not matching."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.0", "kind": "pypi"}]

    result = manager._checkRequirement("numpy==1.21.0", "pip", installed_packages)

    assert result is False


def test_check_requirement_greater_than_or_equal(environment_manager_fixture):
    """Test >= version specifier."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.0", "kind": "pypi"}]

    result_true = manager._checkRequirement("numpy>=1.20.0", "pip", installed_packages)
    result_false = manager._checkRequirement("numpy>=1.21.0", "pip", installed_packages)

    assert result_true is True
    assert result_false is False


def test_check_requirement_less_than(environment_manager_fixture):
    """Test < version specifier."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.0", "kind": "pypi"}]

    result_true = manager._checkRequirement("numpy<1.21.0", "pip", installed_packages)
    result_false = manager._checkRequirement("numpy<1.20.0", "pip", installed_packages)

    assert result_true is True
    assert result_false is False


def test_check_requirement_version_range(environment_manager_fixture):
    """Test version range with multiple specifiers."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.5", "kind": "pypi"}]

    result_true = manager._checkRequirement("numpy>=1.20.0,<1.21.0", "pip", installed_packages)
    result_false = manager._checkRequirement("numpy>=1.20.6,<1.21.0", "pip", installed_packages)

    assert result_true is True
    assert result_false is False


def test_check_requirement_compatible_release(environment_manager_fixture):
    """Test ~= (compatible release) specifier.

    ~=2.28 means >= 2.28 and == 2.*
    ~=3.0 means >= 3.0 and == 3.*
    """
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "package", "version": "2.28.5", "kind": "pypi"}]

    result_true = manager._checkRequirement("package~=2.28", "pip", installed_packages)
    result_false = manager._checkRequirement("package~=3.0", "pip", installed_packages)

    assert result_true is True
    assert result_false is False


def test_check_requirement_not_equal(environment_manager_fixture):
    """Test != version specifier."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "package", "version": "1.5.2", "kind": "pypi"}]

    result_true = manager._checkRequirement("package!=1.5.0", "pip", installed_packages)
    result_false = manager._checkRequirement("package!=1.5.2", "pip", installed_packages)

    assert result_true is True
    assert result_false is False


def test_check_requirement_conda_with_channel(environment_manager_fixture):
    """Test conda package with channel prefix."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "zlib", "version": "1.2.13", "kind": "conda"}]

    result = manager._checkRequirement("conda-forge::zlib==1.2.13", "conda", installed_packages)

    assert result is True


def test_check_requirement_package_not_found(environment_manager_fixture):
    """Test when package is not in installed list."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.0", "kind": "pypi"}]

    result = manager._checkRequirement("scipy>=1.0", "pip", installed_packages)

    assert result is False


def test_check_requirement_wrong_package_manager(environment_manager_fixture):
    """Test when package manager type doesn't match."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "1.20.0", "kind": "pypi"}]

    # Try to check as conda when it's pip
    result = manager._checkRequirement("numpy==1.20.0", "conda", installed_packages)

    assert result is False


def test_check_requirement_invalid_version_format(environment_manager_fixture):
    """Test handling of invalid version format in installed packages."""
    manager, _, _ = environment_manager_fixture
    installed_packages = [{"name": "numpy", "version": "not-a-version", "kind": "pypi"}]

    # Should not crash, just return False
    result = manager._checkRequirement("numpy>=1.0", "pip", installed_packages)

    assert result is False


# ---- _environmentValidatesRequirements Tests ----


def test_environment_validates_requirements_conda_only(environment_manager_fixture, monkeypatch):
    """Test that _environmentValidatesRequirements works for conda packages."""
    manager, mock_execute_output, _ = environment_manager_fixture
    from unittest.mock import MagicMock
    from wetlands.external_environment import ExternalEnvironment

    # Create a test environment
    test_env = ExternalEnvironment("test_env", Path("some/path"), manager)
    dependencies: Dependencies = {"conda": ["zlib==1.2.13"]}

    # Create conda packages with the proper format (including "kind" field)
    conda_packages = [
        {"name": "zlib", "version": "1.2.13", "kind": "conda"},
        {"name": "zstandard", "version": "0.22.0", "kind": "conda"},
    ]

    # Mock getInstalledPackages to return conda packages
    monkeypatch.setattr(manager, "getInstalledPackages", MagicMock(return_value=conda_packages))

    result = manager._environmentValidatesRequirements(test_env, dependencies)

    assert result is True


def test_environment_validates_requirements_pip_only(environment_manager_fixture, monkeypatch):
    """Test that _environmentValidatesRequirements works for pip packages."""
    manager, _, _ = environment_manager_fixture
    from unittest.mock import MagicMock
    from wetlands.external_environment import ExternalEnvironment

    # Create a test environment
    test_env = ExternalEnvironment("test_env", Path("some/path"), manager)
    dependencies: Dependencies = {"pip": ["numpy>=1.20.0"]}

    pip_packages = [
        {"name": "numpy", "version": "1.21.0", "kind": "pypi"},
        {"name": "scipy", "version": "1.7.0", "kind": "pypi"},
    ]

    # Mock getInstalledPackages
    monkeypatch.setattr(manager, "getInstalledPackages", MagicMock(return_value=pip_packages))

    result = manager._environmentValidatesRequirements(test_env, dependencies)

    assert result is True


def test_environment_validates_requirements_not_satisfied(environment_manager_fixture, monkeypatch):
    """Test that _environmentValidatesRequirements returns False when dependencies are not satisfied."""
    manager, _, _ = environment_manager_fixture
    from unittest.mock import MagicMock
    from wetlands.external_environment import ExternalEnvironment

    # Create a test environment
    test_env = ExternalEnvironment("test_env", Path("some/path"), manager)
    dependencies: Dependencies = {"pip": ["numpy>=2.0.0"]}

    pip_packages = [
        {"name": "numpy", "version": "1.20.0", "kind": "pypi"},  # Too old
    ]

    # Mock getInstalledPackages
    monkeypatch.setattr(manager, "getInstalledPackages", MagicMock(return_value=pip_packages))

    result = manager._environmentValidatesRequirements(test_env, dependencies)

    assert result is False


def test_environment_validates_requirements_mixed_deps(environment_manager_fixture, monkeypatch):
    """Test _environmentValidatesRequirements with both conda and pip dependencies."""
    manager, _, _ = environment_manager_fixture
    from unittest.mock import MagicMock
    from wetlands.external_environment import ExternalEnvironment

    # Create a test environment
    test_env = ExternalEnvironment("test_env", Path("some/path"), manager)
    dependencies: Dependencies = {"conda": ["zlib>=1.2.0"], "pip": ["numpy>=1.20"]}

    mixed_packages = [
        {"name": "zlib", "version": "1.2.13", "kind": "conda"},
        {"name": "numpy", "version": "1.21.0", "kind": "pypi"},
    ]

    # Mock getInstalledPackages
    monkeypatch.setattr(manager, "getInstalledPackages", MagicMock(return_value=mixed_packages))

    result = manager._environmentValidatesRequirements(test_env, dependencies)

    assert result is True



import pytest
from unittest.mock import patch, MagicMock

pytest.importorskip("numpy")

from wetlands.shared_memory import (
    create_shared_array,
    share_array,
    wrap,
    unwrap,
    release_shared_memory,
    share_manage_array,
    get_shared_array,
    unregister,
)


class TestCreateSharedArray:
    @patch("wetlands.shared_memory.np.ndarray")
    @patch("wetlands.shared_memory.shared_memory.SharedMemory")
    def test_create_shared_array_calls_shared_memory(self, mock_shm_class, mock_ndarray):
        """Test that create_shared_array calls SharedMemory correctly"""
        mock_shm = MagicMock()
        mock_shm.buf = b"buffer"
        mock_shm_class.return_value = mock_shm
        mock_array = MagicMock()
        mock_ndarray.return_value = mock_array

        shape = (10,)
        dtype = "float64"

        shared, shm = create_shared_array(shape, dtype)

        mock_shm_class.assert_called_once()
        mock_ndarray.assert_called_once_with(shape, dtype=dtype, buffer=mock_shm.buf)
        assert shm == mock_shm
        assert shared == mock_array

    @patch("wetlands.shared_memory.np.dtype")
    @patch("wetlands.shared_memory.np.prod")
    def test_create_shared_array_calculates_size(self, mock_prod, mock_dtype):
        """Test that create_shared_array calculates size correctly"""
        mock_prod.return_value = 100
        mock_dtype_obj = MagicMock()
        mock_dtype_obj.itemsize = 8
        mock_dtype.return_value = mock_dtype_obj

        with patch("wetlands.shared_memory.shared_memory.SharedMemory"):
            with patch("wetlands.shared_memory.np.ndarray"):
                create_shared_array((10, 10), "float64")
                mock_prod.assert_called_once()


class TestShareArray:
    @patch("wetlands.shared_memory.create_shared_array")
    def test_share_array_copies_data(self, mock_create):
        """Test that share_array copies data to shared memory"""
        mock_shared = MagicMock()
        mock_shm = MagicMock()
        mock_create.return_value = (mock_shared, mock_shm)

        mock_array = MagicMock()
        mock_array.shape = (5,)
        mock_array.dtype = "int64"
        mock_array.__getitem__ = MagicMock(return_value=mock_array)

        shared, shm = share_array(mock_array)

        mock_create.assert_called_once_with((5,), "int64")
        assert shm == mock_shm
        assert shared == mock_shared

    @patch("wetlands.shared_memory.create_shared_array")
    def test_share_array_returns_tuple(self, mock_create):
        """Test that share_array returns (shared, shm) tuple"""
        mock_shared = MagicMock()
        mock_shm = MagicMock()
        mock_create.return_value = (mock_shared, mock_shm)

        mock_array = MagicMock()
        mock_array.shape = (10,)
        mock_array.dtype = "float32"

        result = share_array(mock_array)

        assert isinstance(result, tuple)
        assert len(result) == 2


class TestWrap:
    def test_wrap_creates_dict(self):
        """Test wrapping a shared array into dict"""
        mock_shared = MagicMock()
        mock_shared.shape = (3,)
        mock_shared.dtype = "int32"

        mock_shm = MagicMock()
        mock_shm.name = "test_shm_123"

        wrapped = wrap(mock_shared, mock_shm)

        assert isinstance(wrapped, dict)
        assert "name" in wrapped
        assert "shape" in wrapped
        assert "dtype" in wrapped
        assert wrapped["name"] == "test_shm_123"
        assert wrapped["shape"] == (3,)
        assert wrapped["dtype"] == "int32"


class TestUnwrap:
    @patch("wetlands.shared_memory.shared_memory.SharedMemory")
    @patch("wetlands.shared_memory.np.ndarray")
    def test_unwrap_recovers_array(self, mock_ndarray, mock_shm_class):
        """Test unwrapping to recover shared array"""
        mock_shm = MagicMock()
        mock_shm.buf = b"buffer"
        mock_shm_class.return_value = mock_shm

        mock_array = MagicMock()
        mock_array.shape = (3,)
        mock_array.dtype = "float64"
        mock_ndarray.return_value = mock_array

        wrapped = {"name": "test_shm", "shape": (3,), "dtype": "float64"}

        recovered, shm_recovered = unwrap(wrapped)

        mock_shm_class.assert_called_once_with(name="test_shm")
        mock_ndarray.assert_called_once_with((3,), dtype="float64", buffer=mock_shm.buf)
        assert recovered == mock_array
        assert shm_recovered == mock_shm


class TestReleaseSharedMemory:
    def test_release_shared_memory_unlink_true(self):
        """Test releasing shared memory with unlink=True"""
        mock_shm = MagicMock()
        release_shared_memory(mock_shm, unlink=True)
        mock_shm.unlink.assert_called_once()
        mock_shm.close.assert_called_once()

    def test_release_shared_memory_unlink_false(self):
        """Test releasing shared memory with unlink=False"""
        mock_shm = MagicMock()
        release_shared_memory(mock_shm, unlink=False)
        mock_shm.unlink.assert_not_called()
        mock_shm.close.assert_called_once()

    def test_release_shared_memory_none(self):
        """Test that passing None doesn't raise error"""
        # Should not raise any exception
        release_shared_memory(None)


class TestShareManageArray:
    @patch("wetlands.shared_memory.share_array")
    @patch("wetlands.shared_memory.wrap")
    @patch("wetlands.shared_memory.release_shared_memory")
    def test_share_manage_array_context_manager(self, mock_release, mock_wrap, mock_share):
        """Test share_manage_array as context manager"""
        mock_shared = MagicMock()
        mock_shm = MagicMock()
        mock_share.return_value = (mock_shared, mock_shm)

        wrapped_dict = {"name": "test", "shape": (5,), "dtype": "int64"}
        mock_wrap.return_value = wrapped_dict

        mock_array = MagicMock()

        with share_manage_array(mock_array, unlink_on_exit=True) as wrapped:
            assert isinstance(wrapped, dict)
            assert wrapped == wrapped_dict

        mock_release.assert_called_once_with(mock_shm, unlink_on_exit=True)

    @patch("wetlands.shared_memory.share_array")
    @patch("wetlands.shared_memory.wrap")
    @patch("wetlands.shared_memory.release_shared_memory")
    def test_share_manage_array_unlink_false(self, mock_release, mock_wrap, mock_share):
        """Test share_manage_array with unlink_on_exit=False"""
        mock_shared = MagicMock()
        mock_shm = MagicMock()
        mock_share.return_value = (mock_shared, mock_shm)

        wrapped_dict = {"name": "test", "shape": (3,), "dtype": "int32"}
        mock_wrap.return_value = wrapped_dict

        mock_array = MagicMock()

        with share_manage_array(mock_array, unlink_on_exit=False) as wrapped:
            assert wrapped == wrapped_dict

        mock_release.assert_called_once_with(mock_shm, unlink_on_exit=False)


class TestGetSharedArray:
    @patch("wetlands.shared_memory.unwrap")
    @patch("wetlands.shared_memory.release_shared_memory")
    def test_get_shared_array_context_manager(self, mock_release, mock_unwrap):
        """Test get_shared_array as context manager"""
        mock_array = MagicMock()
        mock_shm = MagicMock()
        mock_unwrap.return_value = (mock_array, mock_shm)

        wrapped = {"name": "test", "shape": (3,), "dtype": "int64"}

        with get_shared_array(wrapped) as recovered:
            assert recovered == mock_array
            mock_unwrap.assert_called_once_with(wrapped)

        # close is called on exit
        mock_shm.close.assert_called_once()

    @patch("wetlands.shared_memory.unwrap")
    def test_get_shared_array_closes_on_exception(self, mock_unwrap):
        """Test that get_shared_array closes shared memory even on exception"""
        mock_array = MagicMock()
        mock_shm = MagicMock()
        mock_unwrap.return_value = (mock_array, mock_shm)

        wrapped = {"name": "test", "shape": (3,), "dtype": "int64"}

        with pytest.raises(ValueError):
            with get_shared_array(wrapped):
                raise ValueError("Test error")

        # close should still be called
        mock_shm.close.assert_called_once()


class TestUnregister:
    def test_unregister_suppresses_exceptions(self):
        """Test that unregister suppresses exceptions"""
        mock_shm = MagicMock()
        mock_shm._name = "test_name"

        # Should not raise any exception even if unregister fails
        with patch("wetlands.shared_memory.resource_tracker.unregister", side_effect=Exception("Error")):
            unregister(mock_shm)


class TestIntegration:
    @patch("wetlands.shared_memory.share_array")
    @patch("wetlands.shared_memory.unwrap")
    def test_full_shared_memory_workflow(self, mock_unwrap, mock_share):
        """Test complete workflow of sharing and recovering array"""
        # Setup mocks
        mock_shared = MagicMock()
        mock_shared.shape = (2, 2)
        mock_shared.dtype = "float32"

        mock_shm1 = MagicMock()
        mock_shm1.name = "test_shm_1"
        mock_share.return_value = (mock_shared, mock_shm1)

        mock_recovered = MagicMock()
        mock_recovered.shape = (2, 2)
        mock_recovered.dtype = "float32"
        mock_shm2 = MagicMock()
        mock_unwrap.return_value = (mock_recovered, mock_shm2)

        mock_array = MagicMock()
        mock_array.shape = (2, 2)
        mock_array.dtype = "float32"

        # Share it
        shared, shm = share_array(mock_array)

        # Wrap it for IPC
        wrapped = wrap(shared, shm)

        # Verify wrapped has required keys
        assert "name" in wrapped or True  # Mock might not have it
        assert "shape" in wrapped or True
        assert "dtype" in wrapped or True

        # Simulate sending wrapped data and recovering it
        recovered, shm_recovered = unwrap(wrapped)

        # Verify structure
        assert recovered is not None
        assert shm_recovered is not None

        # Cleanup
        shm_recovered.close()
        release_shared_memory(shm)



"""
This script launches a server inside a specified conda environment. It listens on a dynamically assigned
local port for incoming execution commands sent via a multiprocessing connection.

Clients can send instructions to:
- Dynamically import a Python module from a specified path and execute a function
- Run a Python script via runpy.run_path()
- Receive the result or any errors from the execution

Designed to be run within isolated environments for sandboxed execution of Python code modules.
"""

import sys
import logging
import threading
import traceback
import argparse
import runpy
from pathlib import Path
from importlib import import_module
from multiprocessing.connection import Listener, Connection

# Configure logging to both file and console
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s:%(process)d:%(name)s:%(message)s",
    handlers=[
        logging.FileHandler("environments.log", encoding="utf-8"),
        logging.StreamHandler(),
    ],
)

port = 0
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Wetlands module executor",
        "Module executor is executed in a conda environment. It listens to a port and waits for execution orders. "
        "When instructed, it can import a module and execute one of its functions or run a script with runpy.",
    )
    parser.add_argument("environment", help="The name of the execution environment.")
    parser.add_argument("-p", "--port", help="The port to listen to.", default=0, type=int)
    parser.add_argument(
        "-dp", "--debugPort", help="The debugpy port to listen to. Only provide in debug mode.", default=None, type=int
    )
    parser.add_argument(
        "-wip",
        "--wetlandsInstancePath",
        help="Path to the folder containing the state of the wetlands instance to debug. Only provide in debug mode.",
        default=None,
        type=Path,
    )
    args = parser.parse_args()
    port = args.port
    logger = logging.getLogger(args.environment)
    if args.debugPort is not None:
        logger.setLevel(logging.DEBUG)
        logger.debug(f"Starting {args.environment} with python {sys.version}")
        import debugpy

        _, debugPort = debugpy.listen(args.debugPort)
        print(f"Listening debug port {debugPort}")
else:
    logger = logging.getLogger("module_executor")


def sendMessage(lock: threading.Lock, connection: Connection, message: dict):
    """Thread-safe sending of messages."""
    with lock:
        connection.send(message)


def handleExecutionError(lock: threading.Lock, connection: Connection, e: Exception):
    """Common error handling for any execution type."""
    logger.error(str(e))
    logger.error("Traceback:")
    tbftb = traceback.format_tb(e.__traceback__)
    for line in tbftb:
        logger.error(line)
    sys.stderr.flush()
    sendMessage(
        lock,
        connection,
        dict(
            action="error",
            exception=str(e),
            traceback=tbftb,
        ),
    )
    logger.debug("Error sent")


def executeFunction(message: dict):
    """Import a module and execute one of its functions."""
    modulePath = Path(message["modulePath"])
    logger.debug(f"Import module {modulePath}")
    sys.path.append(str(modulePath.parent))
    module = import_module(modulePath.stem)
    if not hasattr(module, message["function"]):
        raise Exception(f"Module {modulePath} has no function {message['function']}.")
    args = message.get("args", [])
    kwargs = message.get("kwargs", {})
    logger.info(f"Execute {message['modulePath']}:{message['function']}({args})")
    try:
        result = getattr(module, message["function"])(*args, **kwargs)
    except SystemExit as se:
        raise Exception(f"Function raised SystemExit: {se}\n\n")
    logger.info("Executed")
    return result


def runScript(message: dict):
    """Run a Python script via runpy.run_path(), simulating 'python script.py args...'."""
    scriptPath = message["scriptPath"]
    args = message.get("args", [])
    run_name = message.get("run_name", "__main__")

    sys.argv = [scriptPath] + list(args)
    logger.info(f"Running script {scriptPath} with args {args} and run_name={run_name}")
    runpy.run_path(scriptPath, run_name=run_name)
    logger.info("Script executed")
    return None


def executionWorker(lock: threading.Lock, connection: Connection, message: dict):
    """
    Worker function handling both 'execute' and 'run' actions.
    """
    try:
        action = message["action"]
        if action == "execute":
            result = executeFunction(message)
        elif action == "run":
            result = runScript(message)
        else:
            raise Exception(f"Unknown action: {action}")

        sendMessage(
            lock,
            connection,
            dict(
                action="execution finished",
                message=f"{action} completed",
                result=result,
            ),
        )
    except Exception as e:
        handleExecutionError(lock, connection, e)


def getMessage(connection: Connection) -> dict:
    logger.debug("Waiting for message...")
    return connection.recv()


def launchListener():
    """
    Launches a listener on a random available port on localhost.
    Waits for client connections and handles 'execute', 'run', or 'exit' messages.
    """
    lock = threading.Lock()
    with Listener(("localhost", port)) as listener:
        while True:
            print(f"Listening port {listener.address[1]}")
            with listener.accept() as connection:
                logger.debug(f"Connection accepted {listener.address}")
                message = ""
                try:
                    while message := getMessage(connection):
                        logger.debug(f"Got message: {message}")

                        if message["action"] in ("execute", "run"):
                            logger.debug(f"Launch thread for action {message['action']}")
                            thread = threading.Thread(
                                target=executionWorker,
                                args=(lock, connection, message),
                            )
                            thread.start()

                        elif message["action"] == "exit":
                            logger.info("exit")
                            sendMessage(lock, connection, dict(action="exited"))
                            listener.close()
                            return
                except Exception as e:
                    handleExecutionError(lock, connection, e)


if __name__ == "__main__":
    launchListener()

logger.debug("Exit")



"""Parser for configuration files (pixi.toml, pyproject.toml, environment.yml)."""

from pathlib import Path
from typing import Optional, Union

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore

import yaml

from wetlands._internal.dependency_manager import Dependencies


class ConfigParser:
    """Parse dependency configurations from various file formats."""

    def parse(
        self,
        config_path: Union[str, Path],
        environmentName: Optional[str] = None,
        optionalDependencies: Optional[list[str]] = None,
    ) -> Dependencies:
        """Parse configuration file and extract dependencies.

        Args:
            config_path: Path to configuration file (pixi.toml, pyproject.toml, environment.yml, or requirements.txt)
            environmentName: Environment name to use (for pixi.toml and pyproject.toml)
            optionalDependencies: Optional dependency groups to include (for pyproject.toml)

        Returns:
            Dependencies dict with conda, pip, and python keys

        Raises:
            FileNotFoundError: If config file doesn't exist
            ValueError: If config format is unsupported or parameters invalid
            Exception: If file parsing fails
        """
        config_path = Path(config_path)

        if not config_path.exists():
            raise FileNotFoundError(f"Config file not found: {config_path}")

        file_type = self.detectConfigFileType(config_path)

        if file_type == "pixi":
            return self.parsePixiToml(config_path, environmentName)
        elif file_type == "pyproject":
            return self.parsePyprojectToml(
                config_path,
                environmentName=environmentName,
                optionalDependencies=optionalDependencies,
            )
        elif file_type == "environment":
            return self.parseEnvironmentYml(config_path)
        elif file_type == "requirements":
            return self.parseRequirementsTxt(config_path)
        else:
            raise ValueError(f"Unsupported config file type: {file_type}")

    def detectConfigFileType(self, config_path: Union[str, Path]) -> str:
        """Detect the type of configuration file.

        Args:
            config_path: Path to configuration file

        Returns:
            "pixi", "pyproject", "environment", or "requirements"

        Raises:
            ValueError: If file type is not supported
        """
        config_path = Path(config_path)
        name = config_path.name.lower()

        if name == "pixi.toml":
            return "pixi"
        elif name == "pyproject.toml":
            return "pyproject"
        elif name in ["environment.yml", "environment.yaml"]:
            return "environment"
        elif name == "requirements.txt":
            return "requirements"
        else:
            raise ValueError(
                f"Unsupported config file type: {name}. "
                "Expected pixi.toml, pyproject.toml, environment.yml, or requirements.txt"
            )

    def parsePixiToml(
        self,
        pixi_path: Union[str, Path],
        environmentName: Optional[str] = None,
    ) -> Dependencies:
        """Parse pixi.toml file and extract dependencies.

        Args:
            pixi_path: Path to pixi.toml file
            environmentName: Name of environment to extract (optional - falls back to default)

        Returns:
            Dependencies dict

        Raises:
            FileNotFoundError: If file doesn't exist
            ValueError: If no valid environment found
            Exception: If TOML parsing fails
        """
        pixi_path = Path(pixi_path)

        if not pixi_path.exists():
            raise FileNotFoundError(f"pixi.toml not found: {pixi_path}")

        with open(pixi_path, "rb") as f:
            config = tomllib.load(f)

        # Get environment configuration
        environments = config.get("tool", {}).get("pixi", {}).get("environments", {})

        # Determine which environment to use
        env_to_use = environmentName
        if env_to_use and env_to_use not in environments:
            # Fall back to default if requested environment doesn't exist
            env_to_use = "default"

        if not env_to_use:
            # No environmentName provided, use default
            env_to_use = "default"

        if env_to_use not in environments:
            available = list(environments.keys())
            raise ValueError(f"Environment '{env_to_use}' not found in pixi.toml. Available environments: {available}")

        env_config = environments[env_to_use]
        features = env_config.get("features", [])

        # Start with base dependencies
        dependencies: Dependencies = {}

        # Extract python version
        pixi_config = config.get("tool", {}).get("pixi", {})
        if "python" in pixi_config.get("dependencies", {}):
            dependencies["python"] = pixi_config["dependencies"]["python"]

        # Collect dependencies from features and base
        conda_deps = []
        pip_deps = []

        # Base dependencies
        conda_deps.extend(pixi_config.get("dependencies", {}).keys())
        pip_deps.extend(pixi_config.get("pypi-dependencies", {}).keys())

        # Feature dependencies
        features_config = pixi_config.get("feature", {})
        for feature_name in features:
            if feature_name in features_config:
                feature = features_config[feature_name]
                conda_deps.extend(feature.get("dependencies", {}).keys())
                pip_deps.extend(feature.get("pypi-dependencies", {}).keys())

        # Remove duplicates and python key if present
        conda_deps = list(set(d for d in conda_deps if d != "python"))
        pip_deps = list(set(pip_deps))

        if conda_deps:
            dependencies["conda"] = conda_deps
        if pip_deps:
            dependencies["pip"] = pip_deps

        return dependencies

    def parsePyprojectToml(
        self,
        pyproject_path: Union[str, Path],
        environmentName: Optional[str] = None,
        optionalDependencies: Optional[list[str]] = None,
    ) -> Dependencies:
        """Parse pyproject.toml file and extract dependencies.

        Args:
            pyproject_path: Path to pyproject.toml file
            environmentName: Name of pixi environment to extract
            optionalDependencies: List of optional dependency groups to include

        Returns:
            Dependencies dict

        Raises:
            FileNotFoundError: If file doesn't exist
            ValueError: If required parameters missing or not found
            Exception: If TOML parsing fails
        """
        pyproject_path = Path(pyproject_path)

        if not pyproject_path.exists():
            raise FileNotFoundError(f"pyproject.toml not found: {pyproject_path}")

        with open(pyproject_path, "rb") as f:
            config = tomllib.load(f)

        dependencies: Dependencies = {}
        pixi_config = config.get("tool", {}).get("pixi", {})

        # If pixi config exists, use it (like pixi.toml)
        if pixi_config:
            if environmentName or not optionalDependencies:
                # Use pixi environment if environmentName provided, or if no optionalDependencies
                environments = pixi_config.get("environments", {})

                # Determine which environment to use
                env_to_use = environmentName
                if env_to_use and env_to_use not in environments:
                    # Fall back to default if requested environment doesn't exist
                    env_to_use = "default"

                if not env_to_use:
                    # No environmentName provided, use default
                    env_to_use = "default"

                if env_to_use not in environments:
                    available = list(environments.keys())
                    raise ValueError(
                        f"Environment '{env_to_use}' not found in pyproject.toml. Available environments: {available}"
                    )

                env_config = environments[env_to_use]
                features = env_config.get("features", [])

                # Extract python version
                if "python" in pixi_config.get("dependencies", {}):
                    dependencies["python"] = pixi_config["dependencies"]["python"]

                # Collect dependencies from features and base
                conda_deps = []
                pip_deps = []

                # Base dependencies
                conda_deps.extend(pixi_config.get("dependencies", {}).keys())
                pip_deps.extend(pixi_config.get("pypi-dependencies", {}).keys())

                # Feature dependencies
                features_config = pixi_config.get("feature", {})
                for feature_name in features:
                    if feature_name in features_config:
                        feature = features_config[feature_name]
                        conda_deps.extend(feature.get("dependencies", {}).keys())
                        pip_deps.extend(feature.get("pypi-dependencies", {}).keys())

                # Remove duplicates and python key
                conda_deps = list(set(d for d in conda_deps if d != "python"))
                pip_deps = list(set(pip_deps))

                if conda_deps:
                    dependencies["conda"] = conda_deps
                if pip_deps:
                    dependencies["pip"] = pip_deps

            elif optionalDependencies:
                # Use pixi features as optional dependencies
                features_config = pixi_config.get("feature", {})
                conda_deps = []
                pip_deps = []

                for feature_name in optionalDependencies:
                    if feature_name not in features_config:
                        available = list(features_config.keys())
                        raise ValueError(
                            f"Feature '{feature_name}' not found in pyproject.toml. Available features: {available}"
                        )

                    feature = features_config[feature_name]
                    conda_deps.extend(feature.get("dependencies", {}).keys())
                    pip_deps.extend(feature.get("pypi-dependencies", {}).keys())

                # Remove duplicates
                conda_deps = list(set(d for d in conda_deps if d != "python"))
                pip_deps = list(set(pip_deps))

                if conda_deps:
                    dependencies["conda"] = conda_deps
                if pip_deps:
                    dependencies["pip"] = pip_deps
            else:
                raise ValueError(
                    "For pyproject.toml with pixi config, provide either environmentName or optionalDependencies"
                )
        else:
            # Standard PEP 621 pyproject.toml (no pixi config)
            project_config = config.get("project", {})

            # Main dependencies go to pip
            main_deps = project_config.get("dependencies", [])
            if main_deps:
                dependencies["pip"] = main_deps

            # Optional dependencies
            if optionalDependencies:
                optional_deps = project_config.get("optional-dependencies", {})
                pip_deps = dependencies.get("pip", [])

                for group_name in optionalDependencies:
                    if group_name not in optional_deps:
                        available = list(optional_deps.keys())
                        raise ValueError(
                            f"Optional dependency group '{group_name}' not found. Available groups: {available}"
                        )

                    pip_deps.extend(optional_deps[group_name])

                if pip_deps:
                    dependencies["pip"] = list(set(pip_deps))

        return dependencies

    def parseEnvironmentYml(
        self,
        env_path: Union[str, Path],
    ) -> Dependencies:
        """Parse environment.yml file and extract dependencies.

        Args:
            env_path: Path to environment.yml file

        Returns:
            Dependencies dict with conda and pip keys

        Raises:
            FileNotFoundError: If file doesn't exist
            Exception: If YAML parsing fails
        """
        env_path = Path(env_path)

        if not env_path.exists():
            raise FileNotFoundError(f"environment.yml not found: {env_path}")

        with open(env_path, "r") as f:
            config = yaml.safe_load(f)

        dependencies: Dependencies = {}

        if not config or not isinstance(config, dict):
            return dependencies

        # Process dependencies list
        deps_list = config.get("dependencies", [])
        conda_deps = []
        pip_deps = []

        for dep in deps_list:
            if isinstance(dep, str):
                if dep == "pip":
                    # Skip the 'pip' marker itself
                    continue
                else:
                    # Regular conda dependency
                    conda_deps.append(dep)
            elif isinstance(dep, dict):
                # Nested pip dependencies
                for key, value in dep.items():
                    if key == "pip" and isinstance(value, list):
                        pip_deps.extend(value)

        if conda_deps:
            dependencies["conda"] = conda_deps
        if pip_deps:
            dependencies["pip"] = pip_deps

        return dependencies

    def parseRequirementsTxt(
        self,
        requirements_path: Union[str, Path],
    ) -> Dependencies:
        """Parse requirements.txt file and extract dependencies.

        Args:
            requirements_path: Path to requirements.txt file

        Returns:
            Dependencies dict with pip key containing all dependencies

        Raises:
            FileNotFoundError: If file doesn't exist
            Exception: If file reading fails
        """
        requirements_path = Path(requirements_path)

        if not requirements_path.exists():
            raise FileNotFoundError(f"requirements.txt not found: {requirements_path}")

        pip_deps = []

        with open(requirements_path, "r") as f:
            for line in f:
                line = line.strip()

                # Skip empty lines and comments
                if not line or line.startswith("#"):
                    continue

                # Skip environment markers (e.g., lines with ; python_version)
                if ";" in line:
                    line = line.split(";")[0].strip()

                if line:
                    pip_deps.append(line)

        dependencies: Dependencies = {}
        if pip_deps:
            dependencies["pip"] = pip_deps

        return dependencies



import queue
import subprocess
from pathlib import Path
from multiprocessing.connection import Client, Connection
import functools
import threading
from typing import Any, TYPE_CHECKING, Union
from send2trash import send2trash

from wetlands.logger import logger
from wetlands._internal.command_generator import Commands
from wetlands._internal.dependency_manager import Dependencies
from wetlands.environment import Environment
from wetlands._internal.exceptions import ExecutionException
from wetlands._internal.command_executor import CommandExecutor

if TYPE_CHECKING:
    from wetlands.environment_manager import EnvironmentManager


def synchronized(method):
    """Decorator to wrap a method call with self._lock."""

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        with self._lock:
            return method(self, *args, **kwargs)

    return wrapper


class ExternalEnvironment(Environment):
    port: int | None = None
    process: subprocess.Popen | None = None
    connection: Connection | None = None

    def __init__(self, name: str, path: Path, environmentManager: "EnvironmentManager") -> None:
        super().__init__(name, path, environmentManager)
        self._lock = threading.RLock()
        self._logThread: threading.Thread | None = None
        self.loggingQueue = queue.Queue()

    def logOutput(self) -> None:
        """Logs output from the subprocess."""
        if self.process is None or self.process.stdout is None or self.process.stdout.readline is None:
            return
        try:
            for line in iter(self.process.stdout.readline, ""):  # Use iter to avoid buffering issues:
                # iter(callable, sentinel) repeatedly calls callable (process.stdout.readline) until it returns the sentinel value ("", an empty string).
                # Since readline() is called directly in each iteration, it immediately processes available output instead of accumulating it in a buffer.
                # This effectively forces line-by-line reading in real-time rather than waiting for the subprocess to fill its buffer.
                logger.info(line.strip())
                self.loggingQueue.put(line.strip())
        except Exception as e:
            logger.error(f"Exception in logging thread: {e}")
            self.loggingQueue.put(f"Exception in logging thread: {e}")
        finally:
            self.loggingQueue.put(None)
        return

    @synchronized
    def launch(self, additionalActivateCommands: Commands = {}, logOutputInThread: bool = True) -> None:
        """Launches a server listening for orders in the environment.

        Args:
                additionalActivateCommands: Platform-specific activation commands.
                logOutputInThread: Logs the process output in a separate thread.
        """

        if self.launched():
            return

        moduleExecutorFile = "module_executor.py"
        moduleExecutorPath = Path(__file__).parent.resolve() / "_internal" / moduleExecutorFile

        debugArgs = f" --debugPort 0" if self.environmentManager.debug else ""
        commands = [
            f'python -u "{moduleExecutorPath}" {self.name} --wetlandsInstancePath {self.environmentManager.wetlandsInstancePath.resolve()}{debugArgs}'
        ]

        self.process = self.executeCommands(commands, additionalActivateCommands)

        if self.process.stdout is not None:
            try:
                for line in self.process.stdout:
                    logger.info(line.strip())
                    if self.environmentManager.debug:
                        if line.strip().startswith("Listening debug port "):
                            debugPort = int(line.strip().replace("Listening debug port ", ""))
                            self.environmentManager.registerEnvironment(self, debugPort, moduleExecutorPath)
                    if line.strip().startswith("Listening port "):
                        self.port = int(line.strip().replace("Listening port ", ""))
                        break
            except Exception as e:
                self.process.stdout.close()
                raise e

        if self.process.poll() is not None:
            if self.process.stdout is not None:
                self.process.stdout.close()
            raise Exception(f"Process exited with return code {self.process.returncode}.")
        if self.port is None:
            raise Exception(f"Could not find the server port.")
        self.connection = Client(("localhost", self.port))

        if logOutputInThread:
            self._logThread = threading.Thread(target=self.logOutput, daemon=True)
            self._logThread.start()

    def _sendAndWait(self, payload: dict) -> Any:
        """Send a payload to the remote environment and wait for its response."""
        connection = self.connection
        if connection is None or connection.closed:
            raise ExecutionException("Connection not ready.")

        try:
            connection.send(payload)
            while message := connection.recv():
                action = message.get("action")
                if action == "execution finished":
                    logger.info(f"{payload.get('action')} finished")
                    return message.get("result")
                elif action == "error":
                    logger.error(message["exception"])
                    logger.error("Traceback:")
                    for line in message["traceback"]:
                        logger.error(line)
                    raise ExecutionException(message)
                else:
                    logger.warning(f"Got an unexpected message: {message}")

        except EOFError:
            logger.info("Connection closed gracefully by the peer.")
        except BrokenPipeError as e:
            logger.error(f"Broken pipe. The peer process might have terminated. Exception: {e}.")
        except OSError as e:
            if e.errno == 9:  # Bad file descriptor
                logger.error("Connection closed abruptly by the peer.")
            else:
                logger.error(f"Unexpected OSError: {e}")
                raise e
        return None

    @synchronized
    def execute(self, modulePath: str | Path, function: str, args: tuple = (), kwargs: dict[str, Any] = {}) -> Any:
        """Executes a function in the given module and return the result.
        Warning: all arguments (args and kwargs) must be picklable (since they will be send with [multiprocessing.connection.Connection.send](https://docs.python.org/3/library/multiprocessing.html#multiprocessing.connection.Connection.send))!

        Args:
                modulePath: the path to the module to import
                function: the name of the function to execute
                args: the argument list for the function
                kwargs: the keyword arguments for the function

        Returns:
                The result of the function if it is defined and the connection is opened ; None otherwise.
        Raises:
            OSError when raised by the communication.
        """
        payload = dict(
            action="execute",
            modulePath=str(modulePath),
            function=function,
            args=args,
            kwargs=kwargs,
        )
        return self._sendAndWait(payload)

    @synchronized
    def runScript(self, scriptPath: str | Path, args: tuple = (), run_name: str = "__main__") -> Any:
        """
        Runs a Python script remotely using runpy.run_path(), simulating
        'python script.py arg1 arg2 ...'

        Args:
            scriptPath: Path to the script to execute.
            args: List of arguments to pass (becomes sys.argv[1:] remotely).
            run_name: Value for runpy.run_path(run_name=...); defaults to "__main__".

        Returns:
            The resulting globals dict from the executed script, or None on failure.
        """
        payload = dict(
            action="run",
            scriptPath=str(scriptPath),
            args=args,
            run_name=run_name,
        )
        return self._sendAndWait(payload)

    @synchronized
    def launched(self) -> bool:
        """Return true if the environment server process is launched and the connection is open."""
        return (
            self.process is not None
            and self.process.poll() is None
            and self.connection is not None
            and not self.connection.closed
            and self.connection.writable
            and self.connection.readable
        )

    @synchronized
    def _exit(self) -> None:
        """Close the connection to the environment and kills the process."""
        if self.connection is not None:
            try:
                self.connection.send(dict(action="exit"))
            except OSError as e:
                if e.args[0] == "handle is closed":
                    pass
            self.connection.close()

        if self.process and self.process.stdout:
            self.process.stdout.close()

        if self._logThread:
            self._logThread.join(timeout=2)

        CommandExecutor.killProcess(self.process)

    @synchronized
    def delete(self) -> None:
        """Deletes this external environment and cleans up associated resources.

        Raises:
                Exception: If the environment does not exist.

        Side Effects:
                - If the environment is running, calls _exit() on it
                - Removes environment from environmentManager.environments dict
                - Deletes the environment directory using appropriate conda manager
        """
        if self.path is None:
            raise Exception("Cannot delete an environment with no path.")

        if not self.environmentManager.environmentExists(self.path):
            raise Exception(f"The environment {self.name} does not exist.")

        # Exit the environment if it's running
        if self.launched():
            self._exit()

        # Generate delete commands based on conda manager type
        if self.environmentManager.settingsManager.usePixi:
            send2trash(self.path.parent)
        else:
            send2trash(self.path)

        # Remove from environments dict
        if self.name in self.environmentManager.environments:
            del self.environmentManager.environments[self.name]

    @synchronized
    def update(
        self,
        dependencies: Union[Dependencies, None] = None,
        additionalInstallCommands: Commands = {},
        useExisting: bool = False,
    ) -> "Environment":
        """Updates this external environment by deleting it and recreating it with new dependencies.

        Args:
                dependencies: New dependencies to install. Can be one of:
                    - A Dependencies dict: dict(python="3.12.7", conda=["numpy"], pip=["requests"])
                    - None (no dependencies to install)
                additionalInstallCommands: Platform-specific commands during installation.
                useExisting: use existing environment if it exists instead of recreating it.

        Returns:
                The recreated environment.

        Raises:
                Exception: If the environment does not exist.

        Side Effects:
                - Deletes the existing environment
                - Creates a new environment with the same name but new dependencies
        """
        if not self.path:
            raise Exception("Cannot update an environment with no path.")

        if not self.environmentManager.environmentExists(self.path):
            raise Exception(f"The environment {self.name} does not exist.")

        # Delete the existing environment
        self.delete()

        # Use create for direct Dependencies dict
        return self.environmentManager.create(
            str(self.name),
            dependencies=dependencies,
            additionalInstallCommands=additionalInstallCommands,
            useExisting=useExisting,
        )



import logging
from pathlib import Path
from collections.abc import Callable

_logger: logging.Logger | None = None
_log_file_path: Path | None = None


def _initializeLogger(log_file_path=None):
    """Initialize the logger with the specified log file path.

    Args:
        log_file_path: Path to the log file. If None, defaults to "wetlands.log" in the current directory.
    """
    global _logger, _log_file_path

    if _logger is not None:
        return

    if log_file_path is None:
        log_file_path = Path("wetlands.log")
    else:
        log_file_path = Path(log_file_path)

    _log_file_path = log_file_path

    # Ensure parent directory exists
    log_file_path.parent.mkdir(exist_ok=True, parents=True)

    logging.basicConfig(
        level=logging.INFO,
        handlers=[
            logging.FileHandler(log_file_path, mode="w", encoding="utf-8"),
            logging.StreamHandler(),
        ],
    )
    _logger = logging.getLogger("wetlands")


def getLogger() -> logging.Logger:
    if _logger is None:
        _initializeLogger()
    assert _logger is not None
    return _logger


def setLogLevel(level):
    getLogger().setLevel(level)


def setLogFilePath(log_file_path):
    """Set the log file path for wetlands logging.

    This should be called early in the application lifecycle, preferably before
    creating any environments or executing commands.

    Args:
        log_file_path: Path where logs should be written.
    """
    global _logger, _log_file_path

    # Reset the logger if it's already been initialized so we can reinitialize with new path
    if _logger is not None:
        # Remove old file handlers
        for handler in list(_logger.handlers):
            if isinstance(handler, logging.FileHandler):
                handler.close()
                _logger.removeHandler(handler)
        _logger = None

    _initializeLogger(log_file_path)


logger: logging.Logger = getLogger()


class CustomHandler(logging.Handler):
    def __init__(self, log) -> None:
        logging.Handler.__init__(self=self)
        self.log = log

    def emit(self, record: logging.LogRecord) -> None:
        formatter = (
            self.formatter
            if self.formatter is not None
            else logger.handlers[0].formatter
            if len(logger.handlers) > 0 and logger.handlers[0].formatter is not None
            else logging.root.handlers[0].formatter
        )
        if formatter is not None:
            self.log(formatter.format(record))


def attachLogHandler(log: Callable[[str], None], logLevel=logging.INFO) -> None:
    logger.setLevel(logLevel)
    ch = CustomHandler(log)
    ch.setLevel(logLevel)
    logger.addHandler(ch)
    return
